@@ -29,7 +29,10 @@ class Rule(ABC):
     - interactive graphs
     - implicit passing of arguments to the executing command 
     """
-    
+
+    cache_dir: Path
+    tmp_dir: Path
+    is_setup = False
     rules = {}
 
     def __init__(self, name, **kwargs):
@@ -45,43 +48,55 @@ class Rule(ABC):
                     input={'name': "1"}
                 would result in
                     --name "1"
-                You can force string to be displayed without qoutes usingL
+                You can force string to be displayed without quotes using:
                     input={'name': no_quotes("1")}
         """
         assert name not in self.rules
         self.name = name
         self.execution_time = None
         self.rules[name] = self
-        extra_kwargs = set(kwargs) - {'output', 'input', 'group'}
+        extra_kwargs = set(kwargs) - {'output', 'input', 'group', 'parameters'}
         if extra_kwargs:
             raise Exception(f'Unrecognized keyword arguments to {self.__class__.__name__}: {extra_kwargs}')
         self.arguments = subset_dict_preserving_order(
             kwargs,
-            {'input', 'output'}
+            {'input', 'output', 'parameters'}
         )
 
         self.group = kwargs.get('group', None)
         self.outputs = {}
         self.inputs = {}
+        self.parameters = {}
 
         if 'output' in kwargs:
-            self.has_outputs = True
             output = kwargs['output']
             # todo support lists of positionals
             self.outputs = output if isinstance(output, dict) else {'': output}
-        else:
-            self.has_outputs = False
         if 'input' in kwargs:
-            self.has_inputs = True
             input = kwargs['input']
-            self.inputs = input if isinstance(input, dict) else {'': input} 
-        else:
-            self.has_inputs = False
+            self.inputs = input if isinstance(input, dict) else {'': input}
+        if 'parameters' in kwargs:
+            self.parameters = kwargs['parameters']
+
+    @property
+    def has_inputs(self):
+        return len(self.inputs) != 0
+
+    @property
+    def has_outputs(self):
+        return len(self.outputs) != 0
 
     @abstractmethod
-    def run(self, **kwargs) -> int:
-        pass
-    
+    def run(self, use_cache: bool) -> int:
+        if not self.is_setup:
+            raise ValueError('Please set up the rules class settings with Rule.setup() first!')
+
+    @classmethod
+    def setup(cls, cache_dir: Path, tmp_dir: Path):
+        cls.cache_dir = Path(cache_dir)
+        cls.tmp_dir = Path(tmp_dir)
+        cls.is_setup = True
+
     @abstractmethod
     def to_json(self):
         pass
@@ -90,10 +105,12 @@ class Rule(ABC):
         fragments = [repr(self.name)]
         if self.group:
             fragments.append(f'({self.group})')
-        if self.has_inputs or self.has_inputs:
+        if self.has_inputs or self.has_outputs:
             fragments.append('with')
         if self.has_inputs:
             fragments.append(f'{len(self.inputs)} inputs')
+        if self.has_inputs and self.has_outputs:
+            fragments.append('and')
         if self.has_outputs:
             fragments.append(f'{len(self.outputs)} outputs')
         fragments = ' '.join(fragments)
@@ -147,7 +164,9 @@ class ShellRule(Rule):
             for arguments_group in self.arguments.values()
         })
 
-    def run(self, **kwargs) -> int:
+    def run(self, use_cache=False) -> int:
+        super().run(use_cache)
+
         start_time = time.time()
         status = system(f'{self.command} {self.serialized_arguments}')
         self.execution_time = time.time() - start_time
@@ -199,20 +218,26 @@ def expand_run_magics(notebook):
 
 class NotebookRule(Rule):
 
-    cache_dir = '/tmp/nb_cache/'
     options: None
 
+    @property
+    def output_nb_dir(self) -> Path:
+        return self.tmp_dir / 'out'
+
+    @property
+    def reference_nb_dir(self) -> Path:
+        return self.tmp_dir / 'ref'
+
+    @property
+    def stripped_nb_dir(self) -> Path:
+        return self.tmp_dir / 'stripped'
+
     def __init__(
         self, *args, notebook,
         diff=True,
         deduce_io=True,
         deduce_io_from_data_vault=True,
-        # TODO: make this configurable? ue only relative paths with respect to TemporaryDict?
-        # TODO: this has to be project dependent to avoid conflict caches!
-        # TODO: moving the execution logic to a separate module might be a good idea
-        output_nb_dir='/tmp/nbpipeline/out',
-        reference_nb_store='/tmp/nbpipeline/ref/',
-        stripped_nb_dir='/tmp/nbpipeline/stripped/',
+        execute=True,
         **kwargs
     ):
         """Rule for Jupyter Notebooks
@@ -221,24 +246,25 @@ class NotebookRule(Rule):
             deduce_io: whether to automatically deduce inputs and outputs from the code cells tagged "inputs" and "outputs";
                 local variables defined in the cell will be evaluated and used as inputs or outputs.
                 If you want to generate paths with a helper function for brevity, assign a dict of {variable: path}
-                to __inputs__/__outputs__ in the tagged cell using io.create_paths() helper.
+                to `__inputs__`/`__outputs__` in the tagged cell using `io.create_paths()` helper.
             diff: whether to generate diffs against the current state of the notebook
-
+            deduce_io_from_data_vault: whether to deduce the inputs and outputs from `data_vault` magics
+                (`%vault store` and `%vault import`), see https://github.com/krassowski/data-vault
+            execute: if False, the notebook will note be run; useful to include final "leaf" notebooks
+                which may take too long to run, but are not essential to the overall results
         """
         super().__init__(*args, **kwargs)
         self.todos = []
         self.notebook = notebook
         self.absolute_notebook_path = Path(notebook).absolute()
         self.generate_diff = diff
-        self.output_nb_dir = output_nb_dir
-        self.reference_nb_dir = reference_nb_store
-        self.stripped_nb_dir = stripped_nb_dir
         self.diff = None
         self.text_diff = None
         self.fidelity = None
         self.images = []
         self.headers = []
         self.status = None
+        self.execute = execute
 
         from datetime import datetime, timedelta
         month_ago = (datetime.today() - timedelta(days=30)).timestamp()
@@ -279,11 +305,9 @@ class NotebookRule(Rule):
                                 warn(f'Skipping {line} which was previously stored from this notebook to avoid cycles')
                             else:
                                 self.inputs[(index, var_index)] = import_path
-                        self.has_inputs = True
                     elif isinstance(action, StoreAction):
                         store_path = arguments['in'] + '/' + arguments['store']
                         self.outputs[index] = store_path
-                        self.has_outputs = True
                         stored.add(store_path)
 
     def deduce_io_from_tags(self, io_tags={'inputs', 'outputs'}):
@@ -301,7 +325,6 @@ class NotebookRule(Rule):
             assert not getattr(self, f'has_{io}')
             source = ''.join(cell['source'])
             if f'__{io}__' in source:
-                # print(cell['outputs'])
                 assert len(cell['outputs']) == 1
                 # TODO: search through lists
                 values = cell['outputs'][0]['metadata']
@@ -322,7 +345,6 @@ class NotebookRule(Rule):
                     # TODO: add nice exception or warning
                     raise
             setattr(self, io, values)
-            setattr(self, f'has_{io}', True)
 
     def serialize(self, arguments_group):
         return '-p ' + (' -p '.join(
@@ -335,6 +357,7 @@ class NotebookRule(Rule):
         return ' '.join({
             self.serialize(arguments_group)
             for arguments_group in self.arguments.values()
+            if arguments_group
         })
 
     def outline(self, max_depth=3):
@@ -355,22 +378,24 @@ class NotebookRule(Rule):
                     print(f'Creating path "{path}" for "{name}" output argument')
                     path.mkdir(parents=True, exist_ok=True)
 
-    def run(self, use_cache=True, **kwargs) -> int:
+    def run(self, use_cache=True) -> int:
         """
         Run JupyterNotebook using PaperMill and compare the output with reference using nbdime
 
         Returns: status code from the papermill run (0 if successful)
         """
+        super().run(use_cache)
+
         notebook = self.notebook
         path = Path(notebook)
 
-        output_nb_dir = Path(self.output_nb_dir) / path.parent
+        output_nb_dir = self.output_nb_dir / path.parent
         output_nb_dir.mkdir(parents=True, exist_ok=True)
 
-        reference_nb_dir = Path(self.reference_nb_dir) / path.parent
+        reference_nb_dir = self.reference_nb_dir / path.parent
         reference_nb_dir.mkdir(parents=True, exist_ok=True)
 
-        stripped_nb_dir = Path(self.stripped_nb_dir) / path.parent
+        stripped_nb_dir = self.stripped_nb_dir / path.parent
         stripped_nb_dir.mkdir(parents=True, exist_ok=True)
 
         output_nb = output_nb_dir / path.name
@@ -379,7 +404,7 @@ class NotebookRule(Rule):
 
         md5 = run_command(f'md5sum {str(self.absolute_notebook_path)}').split()[0]
 
-        cache_dir = Path(self.cache_dir) / path.parent
+        cache_dir = self.cache_dir / path.parent
         cache_dir.mkdir(parents=True, exist_ok=True)
 
         cache_nb_file = cache_dir / f'{md5}.json'
@@ -389,7 +414,7 @@ class NotebookRule(Rule):
         if use_cache and cache_nb_file.exists():
             with open(cache_nb_file, 'rb') as f:
                 pickled = pickle.load(f)
-                print('Reusing cached results:')
+                print(f'Reusing cached results for {self}')
                 for key in to_cache:
                     setattr(self, key, pickled[key])
                 return 0
@@ -419,16 +444,24 @@ class NotebookRule(Rule):
         # strip outputs (otherwise if it stops, the diff will be too optimistic)
         system(f'cat {self.absolute_notebook_path} | nbstripout > {stripped_nb}')
 
-        # execute
-        start_time = time.time()
-        status = system(f'papermill {stripped_nb} {output_nb} {self.serialized_arguments}') or 0
-        self.execution_time = time.time() - start_time
-        self.status = status
+        if self.execute:
+            # execute
+            start_time = time.time()
+            status = system(f'papermill {stripped_nb} {output_nb} {self.serialized_arguments}') or 0
+            self.execution_time = time.time() - start_time
+        else:
+            status = 0
+            warn(f'Skipping {self} (execute != True)')
 
-        # inject parameters to a "reference" copy (so that we do not have spurious noise in the diff)
-        system(f'papermill {self.absolute_notebook_path} {reference_nb} {self.serialized_arguments} --prepare-only')
+        if self.execute and self.generate_diff:
+
+            # inject parameters to a "reference" copy (so that we do not have spurious noise in the diff)
+            system(
+                f'papermill {self.absolute_notebook_path} {reference_nb} {self.serialized_arguments} --prepare-only'
+                # do not print "Input Notebook:" and "Output Notebook:" for the second time
+                ' --log-level WARNING'
+            )
 
-        if self.generate_diff:
             with NamedTemporaryFile(delete=False) as tf:
                 command = f'nbdiff {reference_nb} {output_nb} --ignore-metadata --ignore-details --out {tf.name}'
                 run_command(command)
@@ -454,6 +487,9 @@ class NotebookRule(Rule):
                     key: getattr(self, key)
                     for key in to_cache
                 }, f)
+
+        self.status = status
+
         return status
 
     def to_json(self):
@@ -506,22 +542,3 @@ class NotebookRule(Rule):
                 </table>>"""
             }
         }
-
-
-"""
-maybe something nested like
-
-class Group:
-    pass
-
-class GroupSharingDirectory(Group):
-    pass
-
-GroupSharingDirectory(
-    name='analyses',
-    members=[
-        NotebookRule(),
-        NotebookRule(),
-    ]
-)
-"""
